/*! \file warpphy.c
\brief Provide PHY-specific functions for interfacing WARPMAC (and MAC code) to the OFDM PHY peripherals

@version 3.0 (SVN rev 724)
@author Chris Hunter & Patrick Murphy

Many of the register names and bit assignments in this file depend on the design of the OFDM PHY. You must
use matching versions of WARPMAC, WARPPHY and ofdm_txrx_mimo. Refer to the wireless reference designs for
known-good combinations of versions.
*/

//WARPPHY Interface
//v3.0 (SVN rev 720)
/***************CHANGELOG*****************
-Cleaned up the code and added tons of comments
-Updated OFDM functions with new registers (matching OFDM rev715)
******************************************/

//XPS-generated header with peripheral parameters
#include "xparameters.h"

//Header for WARPPHY Interface
#include "warpphy.h"

//Header files for WARP peripheral drivers
#include "ofdm_txrx_mimo.h"
#include "ofdm_AGC_mimo.h"
#include "ofdm_timer.h"
#include "ofdm_pktDetector_mimo.h"
#include "EEPROM.h"
#include "radio_controller_basic.h"
#include "radio_controller_ext.h"
#include "radio_controller_adv.h"
#include "radio_controller_5ghz.h"

//********Globals********
unsigned char interruptType = 0x00;//No interrupts by default; user code must enable them later

////Globals for PHY Debugging////
int agctarget = -14;
int agcnoiseEst = -95;
unsigned int sisoMode = 1;
unsigned int numTrainingSyms = 2;
unsigned int pktDly = 58;//251;
unsigned char globalReset = 0;
unsigned int numBaseRate = 2;

unsigned int bothRadios = FIRST_RADIO | SECOND_RADIO;
unsigned int activeRadio = FIRST_RADIO;

//Define a bunch of cryptic constants for the PHY's phase tracking filter
#define INIT_A_KPVAL 0//0x7FFFFFFF//0xA6800
#define INIT_B_KIVAL 0x2500//xD000//x380000//0x7FFFFFFF//0xA6800
#define INIT_B_KPVAL 0xD0000//x2F000//x2620000//x7FFFFFFF//0x54afb0//0x1BC7D10//0x2C3C38//0xA6800
#define INIT_PN_KVAL 0x2700000//xFFFFFFFF
#define INIT_RXFFTOFSET 12
#define TX_FFT_SCALING_STAGE1 1
#define TX_FFT_SCALING_STAGE2 2
#define TX_FFT_SCALING_STAGE3 3

#define RX_FFT_SCALING_STAGE1 0	//1 
#define RX_FFT_SCALING_STAGE2 1	//2
#define RX_FFT_SCALING_STAGE3 1	//1

unsigned int A_KPval = INIT_A_KPVAL;
unsigned int B_KIval = INIT_B_KIVAL;
unsigned int B_KPval = INIT_B_KPVAL;
unsigned int PN_KVal = INIT_PN_KVAL;
unsigned int RxFFT_Window_Offset = INIT_RXFFTOFSET;
/////////////////////////////////


///@brief Initializes the OFDM PHY core
///The OFDM PHY and radio controller cores have many options which must be configured before they can
///be used over the air. This funciton configures these cores with sensible defaults.
int warpphy_init(){

	xil_printf("  Initializing WARPPHY...\r\n");

	///////////////////////////Radio Setup//////////////////////////////
	xil_printf("	Initializing Radio Transmitter...\r\n");
	WarpRadio_v1_Reset((unsigned int *)XPAR_RADIO_CONTROLLER_0_BASEADDR);

	warpphy_applyTxDCOCalibration(FIRST_RADIO);
	warpphy_applyTxDCOCalibration(SECOND_RADIO);

	//Set Tx bandwidth - 0x1 = 24MHz (minimum)
	WarpRadio_v1_TxLpfCornFreqCoarseAdj(0x1, FIRST_RADIO | SECOND_RADIO);

	//Enable hardware control of Tx gains (handled by the radio controller Tx state machine)
	WarpRadio_v1_SoftwareTxGainControl(0, FIRST_RADIO | SECOND_RADIO);
	
	//Four arguments: dly_TxEn, dly_TxStart, dly_GainRampEn, dly_PowerAmpEn, each in units of OPB clock cycles
	WarpRadio_v1_SetTxTiming(FIRST_RADIO | SECOND_RADIO, 0, 250, 100, 0);

	//Three arguments: targetGain, gainStep, stepInterval; max targetGain is 0x3F
	WarpRadio_v1_SetTxGainTiming(FIRST_RADIO | SECOND_RADIO, 0x3F, 0xF, 2);

	WarpRadio_v1_BaseBandTxGain(0x3, FIRST_RADIO | SECOND_RADIO);

	xil_printf("		complete!\r\n");
	
	xil_printf("	Initializing Radio Receiver...");
	//Enable AGC control of receive gains & RxHP
	WarpRadio_v1_RxHpSoftControlDisable(FIRST_RADIO | SECOND_RADIO);
	WarpRadio_v1_SoftwareRxGainControl(0, FIRST_RADIO | SECOND_RADIO);

	//Set bandwith with RxHP=0; RxHighPassCornerFreq(0) is critical for phase noise performance
	WarpRadio_v1_RxHighPassCornerFreq(0, FIRST_RADIO | SECOND_RADIO);

	//Set Rx bandwidth; 0x0 = 15MHz (minimum)
	WarpRadio_v1_RxLpfCornFreqCoarseAdj(0, FIRST_RADIO | SECOND_RADIO);

	xil_printf("complete\r\n");
	/////////////////////////////////////////////////////////////////
	
	////////////////////////////PHY////////////////////////////
	xil_printf("	Initializing OFDM Transmitter...");

	//Disable processing in the Tx PHY to prevent any accidental transmissions during setup
	mimo_ofdmTx_disable();
	mimo_ofdmTx_setFFTScaling((unsigned int)(16*TX_FFT_SCALING_STAGE1 + 4*TX_FFT_SCALING_STAGE2 + 1*TX_FFT_SCALING_STAGE3));

	//Subcarrier indicies for pilot tones - the values here must line up with
	// zeros in the modulation setup for each subcarrier
	mimo_ofdmTx_setPilot1Index((7 + ((64-7)<<16) ));
	mimo_ofdmTx_setPilot2Index((21 + ((64-21)<<16)));

	//Pilot tone values are Fix16_15, so 0x7FFF is +1, 0x8000 is -1
	mimo_ofdmTx_setPilot1Value(0x8000);
	mimo_ofdmTx_setPilot2Value(0x7fff);

	//3 values in this write: number of training symbols, number of base rate symbols and number of full rate symbols
	//The number of full rate symbols is updated with each packet transmission, so it's set to a sensible non-zero value here
	warpphy_setNumSyms(numTrainingSyms + (numBaseRate*256) + (40*65536));

	mimo_ofdmTx_setPreambleScaling(18350);
	mimo_ofdmTx_setControlBits (
		(0 << 4) | //Preamble shift for antenna B
		(TX_SISO_MODE * sisoMode) | //SISO mode
		TX_PILOT_SCRAMBLING | //Pseudo-random scrambling of pilot tones
		//TX_DISABLE_ANTB_PREAMBLE | //Disables preamble on antenna B
		//TX_SISO_ON_ANTB | //Uses antenna B for SISO mode
		0
	);

	xil_printf("complete!\r\n");

	mimo_ofdmTx_enable();

	xil_printf("	Initializing OFDM Receiver...");
	
	//Disable processing in the Rx PHY to prevent any accidental receptions during setup
	mimo_ofdmRx_disable();
	
	mimo_ofdmRx_setFFTScaling((unsigned int)(16*RX_FFT_SCALING_STAGE1 + 4*RX_FFT_SCALING_STAGE2 + 1*RX_FFT_SCALING_STAGE3));
	mimo_ofdmRx_setFFTWindowOffset(INIT_RXFFTOFSET);

	//Scaling value is a 32-bit value, composed of two UFix16_11 values
	// So 0x08000800 scales A/B by exactly 1
	// This value is dependent on the number of training symbols
	mimo_ofdmRx_setRxScaling(0x10001000);

	//Internal packet detector thresholds; ignored when using the external packet detector
	mimo_ofdmRx_setCorrThresh(0x60);
	mimo_ofdmRx_setEnergyThresh(0x2);

	//Long correlator parameters (for packet detection confirmation and symbol timing)
	mimo_ofdmRx_setLongCorrParams(251-16  + (3000*65536)); //251 2500 //4000

	mimo_ofdmRx_setPktDetDly(50);
	mimo_ofdmRx_setNumOFDMSyms(numTrainingSyms + (numBaseRate*65536));
		
	//Three 8-bit values are stored here - each is an index of a byte in the header:
	//bits[15:8]: less-significant byte of pkt length
	//bits[23:16]: more-significant byte of pkt length
	//bits[31:24]: dynamic modulation masks
	mimo_ofdmRx_setByteNums( (unsigned int)((2<<8) + (3<<16) + (0<<24) ));
		
	mimo_ofdmRx_setCFO_B_KI(INIT_B_KIVAL);
	mimo_ofdmRx_setCFO_B_KP(INIT_B_KPVAL);
	
	mimo_ofdmRx_setPNTrack_K(INIT_PN_KVAL);
		
	mimo_ofdmRx_setOptions
	(
		//RESET_BER |
		REQ_LONG_CORR |
		//BIG_PKTBUF_MODE | 
		DYNAMC_PKT_LENGTHS |
		REQ_TWO_LONG_CORR |
		REQ_SHORT_CORR |
		(RX_SISO_MODE * sisoMode) |
		CFO_USE_LONGCORR |
		DEBUG_CFO_OUTSEL | 
		COARSE_CFO_EN | //Enable coarse CFO estimation
		USE_PILOT_ARCTAN |
		//SWITCHING_DIV_EN | //Enable switching diversity at the receiver
		SIMPLE_DYN_MOD_EN |
		EXT_PKT_DETECT |
		//SISO_ON_ANTB |
		//INT_PKT_DETECT,
		//EQ_BYPASS_DIVISION,
		0,
		INTR_BAD_PKTS | INTR_GOOD_PKTS
	);
		
	mimo_ofdmRx_enable();
		
	//Set the modulation schemes:
	//SISO mode, so antenna B is all zeros
	//Base rate at QPSK
	//AntennaA using dynamic modulation, so it's all 1's
	warpphy_set_modulation(2, 0xF, 0x0);

	//For when dynamic modulation is disabled - use QPSK for base-rate and antenna A full-rate
	//warpphy_set_modulation(2, 2, 0x00);

	xil_printf("complete!\r\n");
	///////////////////////////////////////////////////////////
	
	////////////////////////////AGC////////////////////////////
	xil_printf("	Initializing AGC...");
	ofdm_AGC_Initialize(agcnoiseEst);
	ofdm_AGC_setNoiseEstimate(agcnoiseEst);
	ofdm_AGC_FiltSel(0);
	ofdm_AGC_SetDCO(1);
	ofdm_AGC_SetTarget(agctarget);
	ofdm_AGC_Reset();
	xil_printf("complete!\r\n");
	////////////////////////////////////////////////////////////////
	
	////////////////SET UP RSSI PACKET DETECTOR/////////////////////
	xil_printf("	Initializing Packet Detection...");
	ofdm_pktDetector_mimo_WriteReg_csma_enableBusy(PKTDET_BASEADDR, 1);
	ofdm_pktDetector_mimo_WriteReg_csma_enableIdle(PKTDET_BASEADDR, 1);
	ofdm_pktDetector_mimo_WriteReg_pktDet_masterReset(PKTDET_BASEADDR, 1);
	//		ofdm_pktDetector_mimo_WriteReg_pktDet_detectionMask(PKTDET_BASEADDR, 2); //2 = radio slot 3
	ofdm_pktDetector_mimo_WriteReg_pktDet_detectionMask(PKTDET_BASEADDR, 1); //1 = radio slot 2
	ofdm_pktDetector_mimo_WriteReg_pktDet_detectionMode(PKTDET_BASEADDR, 1); //0 = AND, 1 = OR
	ofdm_pktDetector_mimo_WriteReg_pktDet_resetDuration(PKTDET_BASEADDR, 32);
	ofdm_pktDetector_mimo_WriteReg_pktDet_avgLen(PKTDET_BASEADDR, 16);
	ofdm_pktDetector_mimo_WriteReg_pktDet_avgThresh(PKTDET_BASEADDR, 6000); //7000
	ofdm_pktDetector_mimo_WriteReg_csma_avgThresh(PKTDET_BASEADDR, 16000);
	ofdm_pktDetector_mimo_WriteReg_csma_difsPeriod(PKTDET_BASEADDR, 625);
	ofdm_pktDetector_mimo_WriteReg_pktDet_masterReset(PKTDET_BASEADDR, 0);
	ofdm_pktDetector_mimo_WriteReg_pktDet_reset(PKTDET_BASEADDR, 0);
	xil_printf("complete!\r\n");
	////////////////////////////////////////////////////////////////

	//Finally, enable the Rx paths in the radios
	WarpRadio_v1_RxEnable(activeRadio);
	
	warpphy_enableSisoMode();
	warpphy_setNumTrainingSyms(2);
	xil_printf("end of warpphy init...");
	return 0;	
}

///@brief Clears any pending Rx interrupts in the PHY. Was warpphy_pktAck() in previous versions.
///
///The Rx PHY blocks after asserting either its good or bad packet interrupt output. The interrupts
///are cleared by asserting then de-asserting a register bit per interrupt. This funciton clears
///both interrupts; there is no harm in clearing an interrupt that isn't actually asserted.
void warpphy_clearRxInterrupts(){

	//The Rx_PktDone_Reset register has many bits. Only the 2 LSB are used to clear interrupts.
	//The writes below toggle the 2 LSB and retain the values of any other bits.
	ofdm_txrx_mimo_WriteReg_Rx_PktDone_Reset(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_PktDone_Reset(OFDM_BASEADDR) | 0x3);
	ofdm_txrx_mimo_WriteReg_Rx_PktDone_Reset(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_PktDone_Reset(OFDM_BASEADDR) & 0xFFFFFFFC);

	return;
}

///@brief Releases the OFDM Rx PHY master reset
///
///De-asserting RX_GLOBAL_RESET allows the Rx PHY to begin processing packets.
void mimo_ofdmRx_enable(){

	//De-asert the global reset
	ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) & ~RX_GLOBAL_RESET);

	return;
}

///@brief Holds the OFDM Rx PHY in reset
///
///Asserting the RX_GLOBAL_RESET bit clears nearly all the state in the OFDM Rx. Configuration registers are not cleared.
void mimo_ofdmRx_disable(){

	//Assert the global reset
	ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) | RX_GLOBAL_RESET);

	return;
}

///@brief Configures options in the Rx PHY
///
///@param someOptions 32-bit options value, composed on bitwise OR'd values from warpphy.h
///@param intType Selects whether to interrupt on good or bad packets. Bitwise OR'd combination of INTR_BAD_PKTS and  INTR_GOOD_PKTS
void mimo_ofdmRx_setOptions(unsigned int someOptions, unsigned int intType){

	ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, someOptions);

	//Assumes some header has:
	//#define INTR_BAD_PKTS		0x20
	//#define INTR_GOOD_PKTS	0x10
	ofdm_txrx_mimo_WriteReg_Rx_PktDone_Reset(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_PktDone_Reset(OFDM_BASEADDR) | (0x30 & intType));

	return;
}

///@brief Returns the current value of the Rx PHY configuration register
///
/// Returns the value of the OFDM Rx ControlBits register. Use the bit masks from warpphy.h to decode the indivitual bits.
unsigned int mimo_ofdmRx_getOptions(){

	return ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR);
}


///@brief Holds the OFDM Tx in reset
///
/// Holds the OFDM Tx in reset; this prevents any state changes in the Tx PHY. Configuration registers are not affected.
void mimo_ofdmTx_disable(){
	
	//Assert the OFDM Tx master reset and pktDone reset
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, 0x1 | 0x4);

	return;
}

///@brief Releases the OFDM Tx reset
///
/// Releases the OFDM Tx reset
void mimo_ofdmTx_enable(){

	//Assert then clear the OFDM Tx master reset and pktDone reset
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, 0x1 | 0x4);
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, 0x0);

	//Trigger a dummy transmit cycle, so the core appears "done" with the current packet when the MAC first looks
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, 0x2);
	ofdm_txrx_mimo_WriteReg_Tx_Start_Reset_Control(OFDM_BASEADDR, 0x0);

	return;
}


///@brief Initiates the transmission of a packet
///
///Starts the transmission of a packet from the OFDM Tx PHY. If blocking is enabled, this function returns only
///after the transmission finishes. In this mode, the radio receiver is automatically re-enabled. If blocking is disabled,
///the receiver must be re-enabled in user code later.
///
///@param block Selects whether this funciton blocks until the transmission finishes; use TXBLOCK or TXNOBLOCK
///@param radio Selects which radios are enabled for the transmission
void warpphy_pktTx(unsigned int block, unsigned int radio){

	//First checks whether the Tx PHY is already transmitting a packet
	//If so, give up and return
	if(ofdm_txrx_mimo_ReadReg_Tx_PktDone(OFDM_BASEADDR) == 1)
	{
		//warpmac_setDebugGPIO(0x0);
		//----24.9us----

		/* Should there be a check here whether the Rx PHY is receiving a packet?
			And what should happen if it is? It's sort of carrier sensing, but if
			CSMA is disabled, should the Tx pkt be discarded, or held until the Rx
			is no longer busy...
		*/
		WarpRadio_v1_TxEnable(radio);
		usleep(6);//Sleep long enough for the PHY to start transmitting; depends on second argument to SetTxTiming

		if(block == TXBLOCK)
		{
			//Do nothing while the Tx PHY is still busy
			while(ofdm_txrx_mimo_ReadReg_Tx_PktDone(OFDM_BASEADDR) == 0) {}

			//Re-enable the radio receiver
			WarpRadio_v1_RxEnable(radio);
		}
	}
	else xil_printf("ERR: Illegal PHY use\r\n");
	return;
}


///@brief Sets the packet buffer indicies for the OFDM Tx and Rx PHY.
///
///The PLB_BRAM used a the PHY packet buffer is large enough to hold many PHY packets.
///This BRAM is divided into many sub-buffers; the PHY can be set to use any sub-buffer for Tx or Rx.
///
///@param txBufOffset 6-bit integer selecting the sub-buffer for the PHY Tx
///@param rxBufOffset 6-bit integer selecting the sub-buffer for the PHY Rx
void warpphy_setBuffs(unsigned char txBufOffset, unsigned char rxBufOffset){

	//Rx_PktDone_Reset[21:16] compose the Rx buffer offset
	//Rx_PktDone_Reset[13:8]  compose the Tx buffer offset

	//First, zero out the current pkt buff offsets
	ofdm_txrx_mimo_WriteReg_Rx_PktDone_Reset(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_PktDone_Reset(OFDM_BASEADDR) & 0xFF0000FF);

	//Then write the new pkt buff offsets
	ofdm_txrx_mimo_WriteReg_Rx_PktDone_Reset(
		OFDM_BASEADDR,
		ofdm_txrx_mimo_ReadReg_Rx_PktDone_Reset(OFDM_BASEADDR) | 
		( (txBufOffset & 0x3F) << 8 ) |
		( (rxBufOffset & 0x3F) << 16 )
	);

	return;
}

///@brief Configures the PHY for SISO mode
void warpphy_enableSisoMode(){

	//Update the global variable
	sisoMode = 1;
	
	//Enable only 1 radio
	activeRadio = FIRST_RADIO;
	
	//Set SISO mode in the Tx and Rx PHY registers
	mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() | TX_SISO_MODE);
	mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() | RX_SISO_MODE, INTR_BAD_PKTS | INTR_GOOD_PKTS);
	
	return;
}

///@brief Configures the PHY for MIMO mode
void warpphy_enableMimoMode(){

	//Update the global variable
	sisoMode = 0;
	
	//Enable both radios
	activeRadio = activeRadio | SECOND_RADIO;
	
	//Un-set SISO mode in the Tx and Rx PHY registers
	mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() & (~TX_SISO_MODE));
	mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() & (~RX_SISO_MODE), INTR_BAD_PKTS | INTR_GOOD_PKTS);
	
	return;
}

///@brief Sets the number of training symbol periods used per packet
///
///Configures the number of training symbols which are transmitted with each packet. The Tx and Rx nodes must be
///configured for the same number. In SISO mode, the single channel is trained c times. In MIMO mode, each channel
///is trained c/2 times
///
///@param c number of training periods; must be even
void warpphy_setNumTrainingSyms(unsigned int c){
	//Update the global variable; use each time a packet is transmitted
	numTrainingSyms = c;

	//Configure the receiver
	mimo_ofdmRx_setNumOFDMSyms(numTrainingSyms + (numBaseRate*65536));
}

///@brief Configure the flexible modulation/demodulation in the OFDM PHY
///
///The OFDM PHY supports flexible modulation, allowing any combination of schemes per subcarrier
///Currently this code supports simple dynamic modulation, with 48 of 64 subcarriers assigned to carry data, 4 for pilots and 12 empty.
///The modulation scheme in the 48 data subcarriers is set by this funciton.
///
///@param baseRate Modulation scheme for base rate symbols
///@param antAFullRate Modulation scheme for full rate symbols on antenna A
///@param antBFullRate Modulation scheme for full rate symbols on antenna B
void warpphy_set_modulation(unsigned char baseRate, unsigned char antAFullRate, unsigned char antBFullRate)
{
	unsigned char modIndex;
	unsigned int baseRateMask, antAMask, antBMask;
	
	//Define the standard subcarrier mapping - 48 used for data (0xF here), 4 pilots & 12 unused (0x0 here)
	unsigned int modMasks[24] = {0x0FFFFFF0, 0xFFFFFFFF, 0xFF0FFFFF, 0x00000FFF, 0xFF000000, 0xFFFF0FFF, 0xFFFFFFFF, 0xFFFFFF0F, 0x0FFFFFF0, 0xFFFFFFFF, 0xFF0FFFFF, 0x00000FFF, 0xFF000000, 0xFFFF0FFF, 0xFFFFFFFF, 0xFFFFFF0F, 0x0FFFFFF0, 0xFFFFFFFF, 0xFF0FFFFF, 0x00000FFF, 0xFF000000, 0xFFFF0FFF, 0xFFFFFFFF, 0xFFFFFF0F};
	
	switch(baseRate)
	{
		case 1:
			baseRateMask = MODMASK_BPSK;
			break;
		case 2:
			baseRateMask = MODMASK_QPSK;
			break;
		case 4:
			baseRateMask = MODMASK_16QAM;
			break;
		case 6:
			baseRateMask = MODMASK_64QAM;
			break;
		default:
			baseRateMask = MODMASK_QPSK;
			break;
	}
	
	switch(antAFullRate)
	{
		case 1:
			antAMask = MODMASK_BPSK;
			break;
		case 2:
			antAMask = MODMASK_QPSK;
			break;
		case 4:
			antAMask = MODMASK_16QAM;
			break;
		case 6:
			antAMask = MODMASK_64QAM;
			break;
		case 0xF:
			antAMask = 0xFFFFFFFF;
			break;
		default:
			antAMask = MODMASK_QPSK;
			break;
	}
	
	switch(antBFullRate)
	{
		case 1:
			antBMask = MODMASK_BPSK;
			break;
		case 2:
			antBMask = MODMASK_QPSK;
			break;
		case 4:
			antBMask = MODMASK_16QAM;
			break;
		case 6:
			antBMask = MODMASK_64QAM;
			break;
		case 0xF:
			antBMask = 0xFFFFFFFF;
			break;
		default:
			antBMask = MODMASK_QPSK;
			break;
	}
	
	//Configure Tx and Rx antenna A full rate
	for(modIndex=0; modIndex<8; modIndex++)
	{
		*((volatile unsigned int*)(OFDM_BASEADDR+ofdm_txrx_mimo_SMWO_RxModulation_OFFSET+modIndex*sizeof(int))) =
		(modMasks[modIndex] & antAMask);
		
		*((volatile unsigned int*)(OFDM_BASEADDR+ofdm_txrx_mimo_SMWO_TxModulation_OFFSET+modIndex*sizeof(int))) =
			(modMasks[modIndex] & antAMask);
	}
	
	
	//Configure Tx and Rx antenna B full rate
	for(modIndex=8; modIndex<16; modIndex++)
	{
		*((volatile unsigned int*)(OFDM_BASEADDR+ofdm_txrx_mimo_SMWO_RxModulation_OFFSET+modIndex*sizeof(int))) =
		(modMasks[modIndex] & antBMask);
		
		*((volatile unsigned int*)(OFDM_BASEADDR+ofdm_txrx_mimo_SMWO_TxModulation_OFFSET+modIndex*sizeof(int))) =
			(modMasks[modIndex] & antBMask);
	}
	
	//Configure the Tx and Rx base rate
	for(modIndex=16; modIndex<24; modIndex++)
	{
		*((volatile unsigned int*)(OFDM_BASEADDR+ofdm_txrx_mimo_SMWO_RxModulation_OFFSET+modIndex*sizeof(int))) =
		(modMasks[modIndex] & baseRateMask);
		
		*((volatile unsigned int*)(OFDM_BASEADDR+ofdm_txrx_mimo_SMWO_TxModulation_OFFSET+modIndex*sizeof(int))) =
			(modMasks[modIndex] & baseRateMask);
	}
}

///@brief Selects which antenna is used when the PHY is in SISO mode
///
/// The OFDM PHY can use either antenna when in SISO mode. By default, antenna A is used.
///
///@param anteSel 0: use Antenna A, 1: use Antenna B
void warpphy_setSISOAntenna(unsigned char antSel)
{

	if(sisoMode != 1) return; //Not valid if using MIMO mode
	if(mimo_ofdmRx_getOptions()  & SWITCHING_DIV_EN) return; //Not valid if using switching diversity
	
	if(antSel == 0) //choose radio in slot 2
	{
		//Temporarily disable packet detections
		ofdm_pktDetector_mimo_WriteReg_pktDet_reset(PKTDET_BASEADDR, 1);
		
		activeRadio = FIRST_RADIO;
		
		//Switch the pkt detector to listen to just radio #2
		ofdm_pktDetector_mimo_WriteReg_pktDet_detectionMask(PKTDET_BASEADDR, 1);
		
		//Disable antB selection in the PHY
		mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() & (~TX_SISO_ON_ANTB) );
		mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() & (~SISO_ON_ANTB), INTR_BAD_PKTS | INTR_GOOD_PKTS);
		
		//Turn packet detection back on
		ofdm_pktDetector_mimo_WriteReg_pktDet_reset(PKTDET_BASEADDR, 0);
	}
	else if(antSel == 1) //chose radio in slot 3
	{
		//Temporarily disable packet detections
		ofdm_pktDetector_mimo_WriteReg_pktDet_reset(PKTDET_BASEADDR, 1);
		
		activeRadio = SECOND_RADIO;
		
		//Switch the pkt detector to listen to just radio #3
		ofdm_pktDetector_mimo_WriteReg_pktDet_detectionMask(PKTDET_BASEADDR, 2);
		
		//Disable antB selection in the PHY
		mimo_ofdmTx_setControlBits(mimo_ofdmTx_getOptions() | (TX_SISO_ON_ANTB) );
		mimo_ofdmRx_setOptions(mimo_ofdmRx_getOptions() | (SISO_ON_ANTB), INTR_BAD_PKTS | INTR_GOOD_PKTS);
		
		//Turn packet detection back on
		ofdm_pktDetector_mimo_WriteReg_pktDet_reset(PKTDET_BASEADDR, 0);
	}
	
	return;
}

///@brief Set the center frequency of the radio transceivers
///@param band Selects 2.4GHz or 5GHz bands (using GHZ_2 or GHZ_5)
///@param c Selects the channel number in the chosen band
///@return Returns -1 if an invalid band or channel was specified; otherwise returns the new center frequency in MHz
int warpphy_setChannel(unsigned char band, unsigned int c){

	int newFreq = -1;

	if (band == GHZ_2){
		newFreq = WarpRadio_v1_SetCenterFreq2GHz(c, FIRST_RADIO | SECOND_RADIO);
	}
	if (band == GHZ_5){
		newFreq = WarpRadio_v1_SetCenterFreq5GHz(c, FIRST_RADIO | SECOND_RADIO);
	}

	return newFreq;
}

/*********************************************************************/
/* A whole bunch of debugging functions - these will go away someday */
/*********************************************************************/

void warpphy_setPktDlyPlus(){
	pktDly++;
	xil_printf("+PktDly = %d\r\n",pktDly);
	mimo_ofdmRx_setPktDetDly(pktDly);
}
void warpphy_setPktDlyMinus(){
	pktDly--;
	xil_printf("-PktDly = %d\r\n",pktDly);
	mimo_ofdmRx_setPktDetDly(pktDly);
}

/****************************/
/* Phase noise tracking */
/****************************/
void warpphy_set_PN_KPlus(unsigned int increment){
	PN_KVal=PN_KVal+increment;
	mimo_ofdmRx_setPNTrack_K(PN_KVal);
	xil_printf("K: %x\r\n", PN_KVal);
}

void warpphy_set_PN_KMinus(unsigned int decrement){
	PN_KVal=PN_KVal-decrement;
	mimo_ofdmRx_setPNTrack_K(PN_KVal);
	xil_printf("K: %x\r\n", PN_KVal);
}

/*****************/
/* CFO Constants */
/*****************/

void print_CFO_constants ()
{
	xil_printf("AK: %x\tBP: %x\tBI: %x\r\n", A_KPval, B_KPval, B_KIval);
}

void warpphy_set_B_KPPlus(unsigned int increment){
	B_KPval=B_KPval+increment;
	mimo_ofdmRx_setCFO_B_KP(B_KPval);
	print_CFO_constants();
}

void warpphy_set_B_KPMinus(unsigned int decrement){
	B_KPval=B_KPval-decrement;
	mimo_ofdmRx_setCFO_B_KP(B_KPval);
	print_CFO_constants();
}

void warpphy_set_B_KIPlus(unsigned int increment){
	B_KIval=B_KIval+increment;
	mimo_ofdmRx_setCFO_B_KI(B_KIval);
	print_CFO_constants();
}

void warpphy_set_B_KIMinus(unsigned int decrement){
	B_KIval=B_KIval-decrement;
	mimo_ofdmRx_setCFO_B_KI(B_KIval);
	print_CFO_constants();
}

void warpphy_set_CFODebugOutput(unsigned char outputSel){

	if(outputSel)
		ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) | DEBUG_CFO_OUTSEL);
	else
		ofdm_txrx_mimo_WriteReg_Rx_ControlBits(OFDM_BASEADDR, ofdm_txrx_mimo_ReadReg_Rx_ControlBits(OFDM_BASEADDR) & ~DEBUG_CFO_OUTSEL);

	xil_printf("CFO Debug Output %d\r\n",outputSel);
}


/*********************/
/* Other PHY options */
/*********************/

void warpphy_set_FFTOffset_Plus(){
	RxFFT_Window_Offset++;
	mimo_ofdmRx_setFFTWindowOffset(RxFFT_Window_Offset);
	xil_printf("Rx FFT Offset: %d\r\n", RxFFT_Window_Offset);
}

void warpphy_set_FFTOffset_Minus(){
	RxFFT_Window_Offset--;
	mimo_ofdmRx_setFFTWindowOffset(RxFFT_Window_Offset);
	xil_printf("Rx FFT Offset: %d\r\n", RxFFT_Window_Offset);
}

void warpphy_setNoiseTargetPlus(){
	agcnoiseEst++;
	xil_printf("+Noise Estimate = %d\r\n",agcnoiseEst);
	ofdm_AGC_setNoiseEstimate(agcnoiseEst);
}

void warpphy_setNoiseTargetMinus(){
	agcnoiseEst--;
	xil_printf("-Noise Estimate = %d\r\n",agcnoiseEst);
	ofdm_AGC_setNoiseEstimate(agcnoiseEst);
}

void warpphy_setTargetPlus(){
	agctarget++;
	xil_printf("+agctarget = %d\r\n",agctarget);
	//XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1, sevenSegmentMap(-agctarget%10));
	//XGpio_mSetDataReg(XPAR_LED_7SEGMENT_1_BASEADDR, 1, sevenSegmentMap(-agctarget/10));
	ofdm_AGC_SetTarget(agctarget);
}

void warpphy_setTargetMinus(){
	agctarget=agctarget-1;
	xil_printf("-agctarget = %d\r\n",agctarget);
	//XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1, sevenSegmentMap(-agctarget%10));
	//XGpio_mSetDataReg(XPAR_LED_7SEGMENT_1_BASEADDR, 1, sevenSegmentMap(-agctarget/10));
	ofdm_AGC_SetTarget(agctarget);
}


int warpphy_applyTxDCOCalibration(unsigned int radioSelection)
{
	int eepromStatus = 0;
	short calReadback = 0;
	signed short best_I, best_Q;
	unsigned char radioNum;

	//Radio selection will be 0x11111111, 0x22222222, 0x44444444 or 0x88888888
	// corresponding to radios in slots 1, 2, 3 or 4
	// We need the slot number to initialize the EEPROM
	radioNum = (radioSelection & 0xF) == 1 ? 1 : ( (radioSelection & 0xF) == 2 ? 2 : ( (radioSelection & 0xF) == 4 ? 3 : 4 ) );

	//Select the EEPROM on the current radio board
	eepromStatus = WarpEEPROM_EEPROMSelect((unsigned int*)XPAR_EEPROM_CONTROLLER_BASEADDR, radioNum);

	if(eepromStatus != 0)
	{
		xil_printf("TxDCO: EEPROM error\r\n");
		return -1;
	}

	//Initialize the EEPROM controller
	eepromStatus = WarpEEPROM_Initialize((unsigned int*)XPAR_EEPROM_CONTROLLER_BASEADDR);

	if(eepromStatus != 0)
	{
		xil_printf("TxDCO: EEPROM error\r\n");
		return -1;
	}

	//Read the Tx DCO values
	calReadback = WarpEEPROM_ReadRadioCal((unsigned int*)XPAR_EEPROM_CONTROLLER_BASEADDR, 2, 0);

	//Scale the stored values
	best_I = (signed short)(((signed char)(calReadback & 0xFF))<<1);
	best_Q = (signed short)(((signed char)((calReadback>>8) & 0xFF))<<1);

	xil_printf("TxDCO: Applied values to radio %d - I: %d\tQ: %d\r\n", radioNum, best_I, best_Q);

	//Finally, write the Tx DCO values to the DAC
	WarpRadio_v1_DACOffsetAdj(ICHAN, best_I, radioSelection);
	WarpRadio_v1_DACOffsetAdj(QCHAN, best_Q, radioSelection);

	return 0;
}






